# Copyright 2024 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines text string constants used in the model outputs."""

# Model results text.
MODEL_RESULTS_TITLE = 'Marketing Mix Modeling Report'

MODEL_FIT_CARD_ID = 'model-fit'
MODEL_FIT_CARD_TITLE = 'Model fit'
MODEL_FIT_INSIGHTS_FORMAT = """Model fit is a measure of how well your MMM fits
your current data used to train the model."""

EXPECTED_ACTUAL_OUTCOME_CHART_ID = 'expected-actual-outcome-chart'
EXPECTED_ACTUAL_OUTCOME_CHART_TITLE = 'Expected {outcome} vs. actual {outcome}'
EXPECTED_ACTUAL_OUTCOME_CHART_DESCRIPTION_FORMAT = """Note: The baseline represents the
expected {outcome} without any media execution. The shaded blue area represents
the 90% credible interval."""

PREDICTIVE_ACCURACY_TABLE_ID = 'model-fit-statistics-table-chart'
PREDICTIVE_ACCURACY_TABLE_TITLE = 'Model fit metrics'
PREDICTIVE_ACCURACY_TABLE_DESCRIPTION = """Note: R-squared measures the amount
of variation in the data that is explained by the model. The closer it is to 1,
the better the model fit. MAPE measures the mean absolute percentage difference
between the expected and the actual. The closer it is to 0, the better the
model fit. wMAPE is MAPE weighted by the actual {outcome}."""

CHANNEL_CONTRIB_CARD_ID = 'channel-contrib'
CHANNEL_CONTRIB_CARD_TITLE = 'Channel contribution'
CHANNEL_CONTRIB_INSIGHTS_FORMAT = """Your channel contributions help you
understand what drove your {outcome}. {lead_channels} drove the most overall
{outcome}."""

CHANNEL_DRIVERS_CHART_ID = 'channel-drivers-chart'
CHANNEL_DRIVERS_CHART_TITLE = 'Contribution by baseline and marketing channels'
CHANNEL_DRIVERS_CHART_DESCRIPTION = """Note: This graphic encompasses all of
your {outcome} drivers, but breaks down your marketing {outcome} by the baseline
and all channels."""

SPEND_OUTCOME_CHART_ID = 'spend-outcome-chart'
SPEND_OUTCOME_CHART_TITLE = (
    'Spend and {outcome} contribution by marketing channel'
)
SPEND_OUTCOME_CHART_DESCRIPTION = """Note: Return on investment is calculated by
dividing the {outcome} attributed to a channel by marketing costs."""

OUTCOME_CONTRIBUTION_CHART_ID = 'outcome-contribution-chart'
CONTRIBUTION_CHART_TITLE = 'Contribution by baseline and marketing channels'
OUTCOME_CONTRIBUTION_CHART_DESCRIPTION = """Note: This is a percentage breakdown
of all your {outcome}."""

PERFORMANCE_BREAKDOWN_CARD_ID = 'performance-breakdown'
PERFORMANCE_BREAKDOWN_CARD_TITLE = 'Return on investment'
PERFORMANCE_BREAKDOWN_INSIGHTS_FORMAT = """Your return on investment (ROI) helps
you understand how your marketing activities impacted your business objectives.
{lead_roi_channel} drove the highest ROI at {lead_roi_ratio:.1f}. For every $1
you spent on {lead_roi_channel}, you saw ${lead_roi_ratio:.2f} in revenue.
{lead_effectiveness_channel} had the highest effectiveness, which is your
incremental outcome per media unit. {lead_mroi_channel} had the highest marginal
ROI at {lead_mroi_channel_value:.2f}. {lead_cpik_channel} drove the lowest CPIK
at ${lead_cpik_ratio:.2f}. For every KPI unit, you spent ${lead_cpik_ratio:.2f}.
"""

ROI_EFFECTIVENESS_CHART_ID = 'roi-effectiveness-chart'
ROI_EFFECTIVENESS_CHART_TITLE = 'ROI vs. effectiveness'
ROI_EFFECTIVENESS_CHART_DESCRIPTION = """Note: Effectiveness measures the
incremental outcome generated per impression. A low ROI does not necessarily
imply low media effectiveness; it may result from high media cost, as positioned
in the upper-left corner of the chart. Conversely, a high ROI can coexist with
low media effectiveness and low media costs, as indicated in the bottom-right
corner of the chart. The diagonal section of the chart suggests that the ROI is
primarily influenced by media effectiveness. The size of the bubbles represents
the scale of the media spend."""

ROI_MARGINAL_CHART_ID = 'roi-marginal-chart'
ROI_MARGINAL_CHART_TITLE = 'ROI vs. marginal ROI'
ROI_MARGINAL_CHART_DESCRIPTION = """Note: Marginal ROI measures the additional
return generated for every additional dollar spent. It's an indicator of
efficiency of additional spend. Channels with a high ROI but a low marginal ROI
are likely in the saturation phase, where the initial investments have paid off,
but additional investment does not bring in as much return. Conversely, channels
that have a high ROI and a high marginal ROI perform well and continue to yield
high returns with additional spending. The size of the bubbles represents the
scale of the media spend."""

ROI_CHANNEL_CHART_ID = 'roi-channel-chart'
ROI_CHANNEL_CHART_TITLE_FORMAT = 'ROI by channel {ci}'

CPIK_CHANNEL_CHART_ID = 'cpik-channel-chart'
CPIK_CHANNEL_CHART_TITLE_FORMAT = 'CPIK by channel {ci}'
CPIK_CHANNEL_CHART_DESCRIPTION = """Note: CPIK (cost per incremental KPI) point
estimate is determined by the posterior median, whereas ROI point estimate is
determined by the posterior mean."""

RESPONSE_CURVES_CARD_ID = 'response-curves'
RESPONSE_CURVES_CARD_TITLE = 'Response curves'
RESPONSE_CURVES_INSIGHTS_FORMAT = """Your response curves depict the
relationship between marketing spend and the resulting incremental {outcome}."""
OPTIMAL_FREQUENCY_INSIGHTS_FORMAT = """Your optimal weekly frequency for
{rf_channel} is {opt_freq} to maximize ROI."""

RESPONSE_CURVES_CHART_ID = 'response-curves-chart'
RESPONSE_CURVES_CHART_TITLE = (
    'Response curves by marketing channel {top_channels}'
)
RESPONSE_CURVES_CHART_DESCRIPTION_FORMAT = """Note: The response curves are
constructed based on the historical flighting pattern and present the cumulative
incremental {outcome} from the total media spend over the selected time
period."""

OPTIMAL_FREQUENCY_CHART_ID = 'optimal-frequency-chart'
OPTIMAL_FREQUENCY_CHART_TITLE = '{metric} by weekly average frequency'
OPTIMAL_FREQ_CHART_DESCRIPTION = """Note: Optimal frequency is the recommended
average weekly impressions per user (# impressions / # reached users) that
maximizes ROI. When multiple channels have reach and frequency data, only the
channel with the highest spend will be displayed. The same chart can be viewed
for all other channels as described in "Optimize frequency" in the User
Guide."""


# Budget optimization texts.
OPTIMIZATION_TITLE = 'MMM Optimization Report'

SCENARIO_PLAN_CARD_ID = 'scenario-plan'
SCENARIO_PLAN_CARD_TITLE = 'Optimization scenario'
SCENARIO_PLAN_INSIGHTS_UNIFORM_SPEND_BOUNDS = """These are the estimated results
from a {scenario_type} budget scenario with a channel-level spend constraint of
-{lower_bound}% to +{upper_bound}% of the non-optimized spend."""
SCENARIO_PLAN_INSIGHTS_VARIED_SPEND_BOUNDS = """These are the estimated results
from a {scenario_type} budget scenario based on channel-level spend constraints
of the non-optimized spend."""
SCENARIO_PLAN_INSIGHTS_HISTORICAL_BUDGET = """The non-optimized spend for each
channel is equal to the historical spend during the period from {start_date} to
{end_date}."""
SCENARIO_PLAN_INSIGHTS_NEW_BUDGET = """The non-optimized spend for each channel
equals the total budget allocated based on the proportion of historical spend
for each channel during the period from {start_date} to {end_date}."""


NON_OPTIMIZED_BUDGET_LABEL = 'Non-optimized budget'
OPTIMIZED_BUDGET_LABEL = 'Optimized budget'
FIXED_BUDGET_LABEL = 'Fixed'
FLEXIBLE_BUDGET_LABEL = 'Flexible'
NON_OPTIMIZED_ROI_LABEL = 'Non-optimized ROI'
OPTIMIZED_ROI_LABEL = 'Optimized ROI'
NON_OPTIMIZED_CPIK_LABEL = 'Non-optimized CPIK'
OPTIMIZED_CPIK_LABEL = 'Optimized CPIK'
NON_OPTIMIZED_INC_OUTCOME_LABEL = 'Non-optimized incremental {outcome}'
OPTIMIZED_INC_OUTCOME_LABEL = 'Optimized incremental {outcome}'

BUDGET_ALLOCATION_CARD_ID = 'budget-allocation'
BUDGET_ALLOCATION_CARD_TITLE = 'Recommended budget allocation'
BUDGET_ALLOCATION_INSIGHTS = """You can see the channel-level budget shifts and
the performance lift after optimization."""

SPEND_DELTA_CHART_ID = 'spend-delta-chart'
SPEND_DELTA_CHART_TITLE = 'Change in optimized spend for each channel'
SPEND_DELTA_CHART_INSIGHTS = """Note: Each bar represents the change in
optimized spend for a channel. Negative values indicate a decrease in spend
while positive values indicate an increase from your non-optimized spend."""

SPEND_ALLOCATION_CHART_ID = 'spend-allocation-chart'
SPEND_ALLOCATION_CHART_TITLE = 'Optimized budget allocation'

OUTCOME_DELTA_CHART_ID = 'outcome-delta-chart'
OUTCOME_DELTA_CHART_TITLE = (
    'Optimized incremental {outcome} across all channels'
)
OUTCOME_DELTA_CHART_INSIGHTS_FORMAT = """Note: The "non_optimized" bar presents the
incremental {outcome} before optimization, and the subsequent bars represent the
change in incremental {outcome} for each channel based on the optimal spend
level. The final bar labeled "optimized" shows the total incremental {outcome}
after optimization."""

SPEND_ALLOCATION_TABLE_ID = 'spend-allocation-table'

OPTIMIZED_RESPONSE_CURVES_CARD_ID = 'optimized-response-curves'
OPTIMIZED_RESPONSE_CURVES_CARD_TITLE = 'Response curves with optimized spend'
OPTIMIZED_RESPONSE_CURVES_INSIGHTS_FORMAT = """These response curves depict the
relationship between marketing spend and the resulting incremental {outcome} and
show the optimal spend amount for each channel that maximizes the total
incremental {outcome} while staying within the channel-level spend
constraints."""

OPTIMIZED_RESPONSE_CURVES_CHART_ID = 'optimized-response-curves-chart'
OPTIMIZED_RESPONSE_CURVES_CHART_TITLE = 'Optimized response curves'


# Visualizer-only plot titles.
PRIOR_POSTERIOR_DIST_CHART_TITLE = 'Prior vs Posterior Distributions'
RHAT_BOXPLOT_TITLE = 'R-hat Convergence Diagnostic'
ADSTOCK_DECAY_CHART_TITLE = 'Adstock Decay of Effectiveness Over Time'
HILL_SATURATION_CHART_TITLE = 'Hill Saturation Curves'


# Plot labels.
CHANNEL_LABEL = 'Channel'
SPEND_LABEL = 'Spend'
ROI_LABEL = 'ROI'
CPIK_LABEL = 'CPIK'
KPI_LABEL = 'KPI'
REVENUE_LABEL = 'Revenue'
INC_REVENUE_LABEL = 'Incremental revenue'
INC_OUTCOME_LABEL = 'Incremental outcome'
INC_KPI_LABEL = 'Incremental KPI'
OPTIMIZED_SPEND_LABEL = 'Optimized spend'
NONOPTIMIZED_SPEND_LABEL = 'Non-optimized spend'
CURRENT_SPEND_LABEL = 'Current spend'
RESPONSE_CURVES_LABEL = 'Response curves'
HILL_SHADED_REGION_RF_LABEL = 'Relative Distribution of Average Frequency'
HILL_SHADED_REGION_MEDIA_LABEL = (
    'Relative Distribution of Media Units per Capita'
)
HILL_X_AXIS_MEDIA_LABEL = 'Media Units per Capita'
HILL_X_AXIS_RF_LABEL = 'Average Frequency'
HILL_Y_AXIS_LABEL = 'Hill Saturation Level'
EXPECTED_ROI_LABEL = 'Expected ROI'
OPTIMAL_FREQ_LABEL = 'Optimal Frequency'

# Table contents.
DATASET_LABEL = 'Dataset'
R_SQUARED_LABEL = 'R-squared'
MAPE_LABEL = 'MAPE'
WMAPE_LABEL = 'wMAPE'
TRAINING_DATA_LABEL = 'Training Data'
TESTING_DATA_LABEL = 'Testing Data'
ALL_DATA_LABEL = 'All Data'

# Summary metrics table columns.
PCT_IMPRESSIONS_COL = '% impressions'
PCT_SPEND_COL = '% spend'
PCT_CONTRIBUTION_COL = '% contribution'
INC_OUTCOME_COL = 'incremental outcome'
INC_KPI_COL = 'incremental KPI'



# Copyright 2024 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper functions for Data module unit tests."""

from collections.abc import Sequence
import dataclasses
import datetime
import immutabledict
from meridian import constants as c
from meridian.data import input_data
from meridian.data import load
import numpy as np
import pandas as pd
import xarray as xr

_SAMPLE_START_DATE = datetime.date(2021, 1, 25)


def _sample_names(prefix: str, n_names: int | None) -> list[str] | None:
  """Generates a list of sample names.

  It concatenates the same prefix with consecutive numbers to generate a list
  of strings that can be used as sample names of columns/arrays/etc.
  """
  return [prefix + str(n) for n in range(n_names)] if n_names else None


def sample_geos(
    n_geos: int | None, integer_geos: bool = False
) -> list[str] | list[int] | None:
  """Generates a list of sample geos."""
  if integer_geos:
    return list(range(n_geos)) if n_geos else None
  else:
    return _sample_names('geo_', n_geos)


def _sample_times(
    n_times: int,
    start_date: datetime.date = _SAMPLE_START_DATE,
    date_format: str = c.DATE_FORMAT,
) -> list[str]:
  """Generates sample `time`s."""
  return [
      (start_date + datetime.timedelta(weeks=w)).strftime(date_format)
      for w in range(n_times)
  ]


_REQUIRED_COORDS = immutabledict.immutabledict({
    c.GEO: ['geo_0', 'geo_1'],
    c.TIME: _sample_times(n_times=3),
    c.MEDIA_TIME: _sample_times(n_times=3),
    c.CONTROL_VARIABLE: ['control_0', 'control_1'],
})
_MEDIA_COORDS = immutabledict.immutabledict(
    {c.MEDIA_CHANNEL: ['media_channel_0', 'media_channel_1', 'media_channel_2']}
)
_RF_COORDS = immutabledict.immutabledict(
    {c.RF_CHANNEL: ['rf_channel_0', 'rf_channel_1']}
)

_REQUIRED_DATA_VARS = immutabledict.immutabledict({
    c.KPI: (['geo', 'time'], [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]]),
    c.CONTROLS: (
        ['geo', 'time', 'control_variable'],
        [
            [[2.1, 2.2], [2.3, 2.4], [2.5, 2.6]],
            [[2.11, 2.21], [2.31, 2.41], [2.51, 2.61]],
        ],
    ),
    c.POPULATION: (['geo'], [3.1, 3.1]),
})
_OPTIONAL_DATA_VARS = immutabledict.immutabledict({
    c.REVENUE_PER_KPI: (['geo', 'time'], [[1.1, 1.2, 1.3], [1.4, 1.5, 1.6]]),
})
_MEDIA_DATA_VARS = immutabledict.immutabledict({
    c.MEDIA: (
        ['geo', 'media_time', 'media_channel'],
        [
            [[4.1, 4.2, 4.3], [4.4, 4.5, 4.6], [4.7, 4.8, 4.9]],
            [[4.11, 4.21, 4.31], [4.41, 4.51, 4.61], [4.71, 4.81, 4.91]],
        ],
    ),
    c.MEDIA_SPEND: (
        ['geo', 'time', 'media_channel'],
        [
            [[5.1, 5.2, 5.3], [5.4, 5.5, 5.6], [5.7, 5.8, 5.9]],
            [[5.11, 5.21, 5.31], [5.41, 5.51, 5.61], [5.71, 5.81, 5.91]],
        ],
    ),
})
_RF_DATA_VARS = immutabledict.immutabledict({
    c.REACH: (
        ['geo', 'media_time', 'rf_channel'],
        [
            [[6.1, 6.2], [6.3, 6.4], [6.5, 6.6]],
            [[6.11, 6.21], [6.31, 6.41], [6.51, 6.61]],
        ],
    ),
    c.FREQUENCY: (
        ['geo', 'media_time', 'rf_channel'],
        [
            [[7.1, 7.2], [7.3, 7.4], [7.5, 7.6]],
            [[7.11, 7.21], [7.31, 7.41], [7.51, 7.61]],
        ],
    ),
    c.RF_SPEND: (
        ['geo', 'time', 'rf_channel'],
        [
            [[8.1, 8.2], [8.3, 8.4], [8.5, 8.8]],
            [[8.11, 8.21], [8.31, 8.41], [8.51, 8.61]],
        ],
    ),
})

WRONG_DATASET_WO_MEDIA_WO_RF = xr.Dataset(
    coords=_REQUIRED_COORDS,
    data_vars=_REQUIRED_DATA_VARS,
)
WRONG_DATASET_PARTIAL_MEDIA_WO_RF = xr.Dataset(
    coords=_REQUIRED_COORDS,
    data_vars=_REQUIRED_DATA_VARS
    | {
        c.MEDIA: (
            ['geo', 'media_time', 'media_channel'],
            [
                [[4.1, 4.2, 4.3], [4.4, 4.5, 4.6], [4.7, 4.8, 4.9]],
                [[4.11, 4.21, 4.31], [4.41, 4.51, 4.61], [4.71, 4.81, 4.91]],
            ],
        ),
    },
)
WRONG_DATASET_WO_MEDIA_PARTIAL_RF = xr.Dataset(
    coords=_REQUIRED_COORDS,
    data_vars=_REQUIRED_DATA_VARS
    | {
        c.REACH: (
            ['geo', 'media_time', 'rf_channel'],
            [
                [[6.1, 6.2], [6.3, 6.4], [6.5, 6.6]],
                [[6.11, 6.21], [6.31, 6.41], [6.51, 6.61]],
            ],
        ),
        c.RF_SPEND: (
            ['geo', 'time', 'rf_channel'],
            [
                [[8.1, 8.2], [8.3, 8.4], [8.5, 8.8]],
                [[8.11, 8.21], [8.31, 8.41], [8.51, 8.61]],
            ],
        ),
    },
)
WRONG_DATASET_PARTIAL_MEDIA_PARTIAL_RF = xr.Dataset(
    coords=_REQUIRED_COORDS,
    data_vars=_REQUIRED_DATA_VARS
    | {
        c.MEDIA_SPEND: (
            ['geo', 'time', 'media_channel'],
            [
                [[5.1, 5.2, 5.3], [5.4, 5.5, 5.6], [5.7, 5.8, 5.9]],
                [[5.11, 5.21, 5.31], [5.41, 5.51, 5.61], [5.71, 5.81, 5.91]],
            ],
        ),
    }
    | {
        c.REACH: (
            ['geo', 'media_time', 'rf_channel'],
            [
                [[6.1, 6.2], [6.3, 6.4], [6.5, 6.6]],
                [[6.11, 6.21], [6.31, 6.41], [6.51, 6.61]],
            ],
        ),
        c.RF_SPEND: (
            ['geo', 'time', 'rf_channel'],
            [
                [[8.1, 8.2], [8.3, 8.4], [8.5, 8.8]],
                [[8.11, 8.21], [8.31, 8.41], [8.51, 8.61]],
            ],
        ),
    },
)
WRONG_DATASET_W_MEDIA_PARTIAL_RF = xr.Dataset(
    coords=_REQUIRED_COORDS | _MEDIA_COORDS,
    data_vars=_REQUIRED_DATA_VARS
    | _MEDIA_DATA_VARS
    | {
        c.REACH: (
            ['geo', 'media_time', 'rf_channel'],
            [
                [[6.1, 6.2], [6.3, 6.4], [6.5, 6.6]],
                [[6.11, 6.21], [6.31, 6.41], [6.51, 6.61]],
            ],
        ),
        c.RF_SPEND: (
            ['geo', 'time', 'rf_channel'],
            [
                [[8.1, 8.2], [8.3, 8.4], [8.5, 8.8]],
                [[8.11, 8.21], [8.31, 8.41], [8.51, 8.61]],
            ],
        ),
    },
)
WRONG_DATASET_PARTIAL_MEDIA_W_RF = xr.Dataset(
    coords=_REQUIRED_COORDS | _RF_COORDS,
    data_vars=_REQUIRED_DATA_VARS
    | _RF_DATA_VARS
    | {
        c.MEDIA: (
            ['geo', 'media_time', 'media_channel'],
            [
                [[4.1, 4.2, 4.3], [4.4, 4.5, 4.6], [4.7, 4.8, 4.9]],
                [[4.11, 4.21, 4.31], [4.41, 4.51, 4.61], [4.71, 4.81, 4.91]],
            ],
        ),
    },
)

DATASET_WITHOUT_GEO_VARIATION_IN_CONTROLS = xr.Dataset(
    coords=_REQUIRED_COORDS | _MEDIA_COORDS | _RF_COORDS,
    data_vars=_MEDIA_DATA_VARS
    | _RF_DATA_VARS
    | {
        c.KPI: (['geo', 'time'], [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]]),
        c.REVENUE_PER_KPI: (
            ['geo', 'time'],
            [[1.1, 1.2, 1.3], [1.4, 1.5, 1.6]],
        ),
        c.CONTROLS: (
            ['geo', 'time', 'control_variable'],
            [
                [[2.1, 2.2], [2.3, 2.4], [2.5, 2.6]],
                [[2.1, 2.2], [2.3, 2.4], [2.5, 2.6]],
            ],
        ),
        c.POPULATION: (['geo'], [3.1, 3.1]),
    },
)

DATASET_WITHOUT_GEO_VARIATION_IN_MEDIA = xr.Dataset(
    coords=_REQUIRED_COORDS | _MEDIA_COORDS | _RF_COORDS,
    data_vars=_REQUIRED_DATA_VARS
    | _OPTIONAL_DATA_VARS
    | _RF_DATA_VARS
    | {
        c.MEDIA: (
            ['geo', 'media_time', 'media_channel'],
            [
                [[4.1, 4.2, 4.3], [4.4, 4.5, 4.6], [4.7, 4.8, 4.9]],
                [[4.11, 4.2, 4.3], [4.41, 4.5, 4.6], [4.71, 4.8, 4.9]],
            ],
        ),
        c.MEDIA_SPEND: (
            ['geo', 'time', 'media_channel'],
            [
                [[5.1, 5.2, 5.3], [5.4, 5.5, 5.6], [5.7, 5.8, 5.9]],
                [[5.11, 5.21, 5.31], [5.41, 5.51, 5.61], [5.71, 5.81, 5.91]],
            ],
        ),
    },
)

DATASET_WITHOUT_GEO_VARIATION_IN_REACH = xr.Dataset(
    coords=_REQUIRED_COORDS | _MEDIA_COORDS | _RF_COORDS,
    data_vars=_REQUIRED_DATA_VARS
    | _OPTIONAL_DATA_VARS
    | _MEDIA_DATA_VARS
    | {
        c.REACH: (
            ['geo', 'media_time', 'rf_channel'],
            [
                [[6.1, 6.2], [6.3, 6.4], [6.5, 6.6]],
                [[6.1, 6.2], [6.3, 6.4], [6.5, 6.6]],
            ],
        ),
        c.FREQUENCY: (
            ['geo', 'media_time', 'rf_channel'],
            [
                [[7.1, 7.2], [7.3, 7.4], [7.5, 7.6]],
                [[7.11, 7.21], [7.31, 7.41], [7.51, 7.61]],
            ],
        ),
        c.RF_SPEND: (
            ['geo', 'time', 'rf_channel'],
            [
                [[8.1, 8.2], [8.3, 8.4], [8.5, 8.8]],
                [[8.11, 8.21], [8.31, 8.41], [8.51, 8.61]],
            ],
        ),
    },
)

DATASET_WITHOUT_TIME_VARIATION_IN_CONTROLS = xr.Dataset(
    coords=_REQUIRED_COORDS | _MEDIA_COORDS | _RF_COORDS,
    data_vars=_MEDIA_DATA_VARS
    | _RF_DATA_VARS
    | {
        c.KPI: (['geo', 'time'], [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]]),
        c.REVENUE_PER_KPI: (
            ['geo', 'time'],
            [[1.1, 1.2, 1.3], [1.4, 1.5, 1.6]],
        ),
        c.CONTROLS: (
            ['geo', 'time', 'control_variable'],
            [
                [[2.1, 2.2], [2.1, 2.2], [2.1, 2.2]],
                [[2.7, 2.8], [2.7, 2.8], [2.7, 2.8]],
            ],
        ),
        c.POPULATION: (['geo'], [3.1, 3.1]),
    },
)

DATASET_WITHOUT_TIME_VARIATION_IN_MEDIA = xr.Dataset(
    coords=_REQUIRED_COORDS | _MEDIA_COORDS | _RF_COORDS,
    data_vars=_REQUIRED_DATA_VARS
    | _OPTIONAL_DATA_VARS
    | _RF_DATA_VARS
    | {
        c.MEDIA: (
            ['geo', 'media_time', 'media_channel'],
            [
                [[4.1, 4.2, 4.3], [4.4, 4.2, 4.3], [4.7, 4.2, 4.3]],
                [[4.11, 4.21, 4.31], [4.41, 4.21, 4.31], [4.71, 4.21, 4.31]],
            ],
        ),
        c.MEDIA_SPEND: (
            ['geo', 'time', 'media_channel'],
            [
                [[5.1, 5.2, 5.3], [5.4, 5.5, 5.6], [5.7, 5.8, 5.9]],
                [[5.11, 5.21, 5.31], [5.41, 5.51, 5.61], [5.71, 5.81, 5.91]],
            ],
        ),
    },
)

DATASET_WITHOUT_TIME_VARIATION_IN_REACH = xr.Dataset(
    coords=_REQUIRED_COORDS | _MEDIA_COORDS | _RF_COORDS,
    data_vars=_REQUIRED_DATA_VARS
    | _OPTIONAL_DATA_VARS
    | _MEDIA_DATA_VARS
    | {
        c.REACH: (
            ['geo', 'media_time', 'rf_channel'],
            [
                [[6.1, 6.2], [6.1, 6.2], [6.1, 6.2]],
                [[6.7, 6.8], [6.7, 6.8], [6.7, 6.8]],
            ],
        ),
        c.FREQUENCY: (
            ['geo', 'media_time', 'rf_channel'],
            [
                [[7.1, 7.2], [7.3, 7.4], [7.5, 7.6]],
                [[7.11, 7.21], [7.31, 7.41], [7.51, 7.61]],
            ],
        ),
        c.RF_SPEND: (
            ['geo', 'time', 'rf_channel'],
            [
                [[8.1, 8.2], [8.3, 8.4], [8.5, 8.8]],
                [[8.11, 8.21], [8.31, 8.41], [8.51, 8.61]],
            ],
        ),
    },
)

_NATIONAL_COORDS = immutabledict.immutabledict({
    c.TIME: [
        _SAMPLE_START_DATE.strftime(c.DATE_FORMAT),
        (_SAMPLE_START_DATE + datetime.timedelta(weeks=1)).strftime(
            c.DATE_FORMAT
        ),
    ],
    c.CONTROL_VARIABLE: ['control_0', 'control_1'],
    c.MEDIA_TIME: [
        _SAMPLE_START_DATE.strftime(c.DATE_FORMAT),
        (_SAMPLE_START_DATE + datetime.timedelta(weeks=1)).strftime(
            c.DATE_FORMAT
        ),
    ],
    c.MEDIA_CHANNEL: ['media_channel_0'],
})

_NATIONAL_DATA_VARS_W_GEO = immutabledict.immutabledict({
    c.CONTROLS: (
        ['geo', 'time', 'control_variable'],
        [[[0.1, 0.2], [0.3, 0.4]]],
    ),
    c.MEDIA: (['geo', 'media_time', 'media_channel'], [[[1.1], [1.2]]]),
    c.KPI: (['geo', 'time'], [[2.1, 2.2]]),
    c.MEDIA_SPEND: (['geo', 'time', 'media_channel'], [[[3.1], [3.2]]]),
    c.REVENUE_PER_KPI: (['geo', 'time'], [[4.1, 4.2]]),
})
_NATIONAL_DATA_VARS_WO_GEO = immutabledict.immutabledict({
    c.CONTROLS: (['time', 'control_variable'], [[0.1, 0.2], [0.3, 0.4]]),
    c.MEDIA: (['media_time', 'media_channel'], [[1.1], [1.2]]),
    c.KPI: (['time'], [2.1, 2.2]),
    c.MEDIA_SPEND: (['time', 'media_channel'], [[3.1], [3.2]]),
    c.REVENUE_PER_KPI: (['time'], [4.1, 4.2]),
})

NATIONAL_DATASET_W_POPULATION_W_GEO = xr.Dataset(
    coords=_NATIONAL_COORDS | {c.GEO: ['geo_0']},
    data_vars=_NATIONAL_DATA_VARS_W_GEO | {c.POPULATION: (['geo'], [5.1])},
)
NATIONAL_DATASET_W_SCALAR_POPULATION_WO_GEO = xr.Dataset(
    coords=_NATIONAL_COORDS,
    data_vars=_NATIONAL_DATA_VARS_WO_GEO | {c.POPULATION: 5.1},
)
NATIONAL_DATASET_W_SINGLE_POPULATION_WO_GEO = xr.Dataset(
    coords=_NATIONAL_COORDS,
    data_vars=_NATIONAL_DATA_VARS_WO_GEO | {c.POPULATION: 5.1},
)
NATIONAL_DATASET_W_NONE_POPULATION_WO_GEO = xr.Dataset(
    coords=_NATIONAL_COORDS,
    data_vars=_NATIONAL_DATA_VARS_WO_GEO | {c.POPULATION: None},
)
NATIONAL_DATASET_WO_POPULATION_W_GEO = xr.Dataset(
    coords=_NATIONAL_COORDS | {c.GEO: ['geo_0']},
    data_vars=_NATIONAL_DATA_VARS_W_GEO,
)
NATIONAL_DATASET_WO_POPULATION_WO_GEO = xr.Dataset(
    coords=_NATIONAL_COORDS,
    data_vars=_NATIONAL_DATA_VARS_WO_GEO,
)
EXPECTED_NATIONAL_DATASET = xr.Dataset(
    coords=_NATIONAL_COORDS | {c.GEO: [c.NATIONAL_MODEL_DEFAULT_GEO_NAME]},
    data_vars=_NATIONAL_DATA_VARS_W_GEO | {c.POPULATION: (['geo'], [1.0])},
)


NATIONAL_DATA_DICT_WO_POPULATION_WO_GEO = immutabledict.immutabledict({
    'time': [
        _SAMPLE_START_DATE.strftime(c.DATE_FORMAT),
        (_SAMPLE_START_DATE + datetime.timedelta(weeks=1)).strftime(
            c.DATE_FORMAT
        ),
    ],
    'kpi': [0.1, 0.2],
    'revenue_per_kpi': [1.1, 1.2],
    'media_0': [3.1, 3.2],
    'media_1': [3.3, 3.4],
    'media_2': [3.5, 3.6],
    'media_spend_0': [4.1, 4.2],
    'media_spend_1': [4.3, 4.4],
    'media_spend_2': [4.5, 4.6],
    'control_0': [5.1, 5.2],
    'control_1': [5.3, 5.4],
})
NATIONAL_DATA_DICT_W_POPULATION_W_GEO = (
    NATIONAL_DATA_DICT_WO_POPULATION_WO_GEO
    | {
        'geo': 'geo_0',
        'population': 2.1,
    }
)
NATIONAL_DATA_DICT_W_POPULATION_W_GEO_RENAMED = (
    NATIONAL_DATA_DICT_WO_POPULATION_WO_GEO
    | {
        'City': 'geo_0',
        'Population': 2.1,
    }
)
NATIONAL_DATA_DICT_W_POPULATION_WO_GEO = (
    NATIONAL_DATA_DICT_WO_POPULATION_WO_GEO | {'population': 2.1}
)
NATIONAL_DATA_DICT_WO_POPULATION_W_GEO = (
    NATIONAL_DATA_DICT_WO_POPULATION_WO_GEO | {'geo': 'geo_0'}
)
EXPECTED_NATIONAL_DATA_DICT = NATIONAL_DATA_DICT_WO_POPULATION_WO_GEO | {
    'geo': c.NATIONAL_MODEL_DEFAULT_GEO_NAME,
    'population': c.NATIONAL_MODEL_DEFAULT_POPULATION_VALUE,
}

NATIONAL_COORD_TO_COLUMNS_WO_POPULATION_WO_GEO = load.CoordToColumns(
    time='time',
    kpi='kpi',
    revenue_per_kpi='revenue_per_kpi',
    media=['media_0', 'media_1', 'media_2'],
    media_spend=['media_spend_0', 'media_spend_1', 'media_spend_2'],
    controls=['control_0', 'control_1'],
)
NATIONAL_COORD_TO_COLUMNS_W_POPULATION_W_GEO_RENAMED = dataclasses.replace(
    NATIONAL_COORD_TO_COLUMNS_WO_POPULATION_WO_GEO,
    population='Population',
    geo='City',
)
NATIONAL_COORD_TO_COLUMNS_W_POPULATION_W_GEO = dataclasses.replace(
    NATIONAL_COORD_TO_COLUMNS_WO_POPULATION_WO_GEO,
    population='population',
    geo='geo',
)
NATIONAL_COORD_TO_COLUMNS_W_POPULATION_WO_GEO = dataclasses.replace(
    NATIONAL_COORD_TO_COLUMNS_WO_POPULATION_WO_GEO,
    population='population',
)
NATIONAL_COORD_TO_COLUMNS_WO_POPULATION_W_GEO = dataclasses.replace(
    NATIONAL_COORD_TO_COLUMNS_WO_POPULATION_WO_GEO,
    geo='geo',
)


def random_media_da(
    n_geos: int,
    n_times: int,
    n_media_times: int,
    n_media_channels: int,
    seed: int = 0,
    date_format: str = c.DATE_FORMAT,
    explicit_geo_names: Sequence[str] | None = None,
    explicit_time_index: Sequence[str] | None = None,
    explicit_media_channel_names: Sequence[str] | None = None,
    array_name: str = 'media',
    channel_variable_name: str = 'media_channel',
    channel_prefix: str = 'ch_',
    integer_geos: bool = False,
) -> xr.DataArray:
  """Generates a sample `media` DataArray.

  Args:
    n_geos: Number of geos
    n_times: Number of time periods
    n_media_times: Number of media time periods
    n_media_channels: Number of media channels
    seed: Random seed used by `np.random.seed()`
    date_format: The date format to use for time coordinate labels
    explicit_geo_names: If given, ignore `n_geos` and use this as is.
    explicit_time_index: If given, ignore `date_format` and use this as is
    explicit_media_channel_names: If given, ignore `n_media_channels` and use
      this as is
    array_name: The name of the array to be created
    channel_variable_name: The name of the channel variable
    channel_prefix: The prefix of the channel names
    integer_geos: If True, the geos will be integers.

  Returns:
    A DataArray containing random data.
  """

  np.random.seed(seed)

  start_date = _SAMPLE_START_DATE
  if n_times < n_media_times:
    start_date -= datetime.timedelta(weeks=(n_media_times - n_times))

  media = np.round(
      abs(
          np.random.normal(5, 5, size=(n_geos, n_media_times, n_media_channels))
      )
  )
  if explicit_geo_names is None:
    geos = sample_geos(n_geos, integer_geos)
  else:
    geos = explicit_geo_names

  if explicit_time_index is None:
    media_time = _sample_times(
        n_times=n_media_times,
        start_date=start_date,
        date_format=date_format,
    )
  else:
    media_time = explicit_time_index

  media_channels = (
      explicit_media_channel_names
      if explicit_media_channel_names is not None
      else _sample_names(prefix=channel_prefix, n_names=n_media_channels)
  )
  return xr.DataArray(
      media,
      dims=['geo', 'media_time', channel_variable_name],
      coords={
          'geo': geos,
          'media_time': media_time,
          channel_variable_name: media_channels,
      },
      name=array_name,
  )


def random_organic_media_da(
    n_geos: int,
    n_times: int,
    n_media_times: int,
    n_organic_media_channels: int,
    seed: int = 0,
    date_format: str = c.DATE_FORMAT,
    explicit_time_index: Sequence[str] | None = None,
    explicit_media_channel_names: Sequence[str] | None = None,
    integer_geos: bool = False,
) -> xr.DataArray:
  """Generates a sample `organic_media` DataArray."""
  return random_media_da(
      n_geos=n_geos,
      n_times=n_times,
      n_media_times=n_media_times,
      n_media_channels=n_organic_media_channels,
      seed=seed,
      date_format=date_format,
      explicit_time_index=explicit_time_index,
      explicit_media_channel_names=explicit_media_channel_names,
      array_name='organic_media',
      channel_variable_name='organic_media_channel',
      channel_prefix='organic_media_',
      integer_geos=integer_geos,
  )


def random_media_spend_nd_da(
    n_geos: int | None = None,
    n_times: int | None = None,
    n_media_channels: int | None = None,
    seed=0,
    integer_geos: bool = False,
) -> xr.DataArray:
  """Generates a sample N-dimensional `media_spend` DataArray.

  This function generates a 1-D, 2-D or 3-D version of the `media_spend`
  DataArray depending on the `n_geos`, `n_times` and `n_media_channels`
  arguments. There are only 2 shapes accepted by the `InputData` class:
  `(geo, time, media_channel)` and `(media_channel)` but we use also the 2-D
  version of this function to test if `InputData` fails to initialize with a 2-D
  media_spend data.

  Args:
    n_geos: Number of geos in the created `media_spend` array or `None` if the
      created `media_spend` should not contain the `geo` dimension.
    n_times: Number of time periods in the created `media_spend` array or `None`
      if the created array should not contain the `time` dimension.
    n_media_channels: Number of channels in the created `media_spend` array.
    seed: Random seed used by `np.random.seed()`.
    integer_geos: If True, the geos will be integers.

  Returns:
    A DataArray containing the generated `media_spend` data with the given
    dimensions.
  """
  np.random.seed(seed)

  dims = []
  coords = {}
  if n_geos is not None:
    dims.append('geo')
    coords['geo'] = sample_geos(n_geos, integer_geos)
  if n_times is not None:
    dims.append('time')
    coords['time'] = _sample_times(n_times=n_times)
  if n_media_channels is not None:
    dims.append('media_channel')
    coords['media_channel'] = _sample_names(
        prefix='ch_', n_names=n_media_channels
    )

  if dims == ['geo', 'time', 'media_channel']:
    shape = (n_geos, n_times, n_media_channels)
  elif dims == ['media_channel']:
    shape = (n_media_channels,)
  elif dims == ['geo', 'media_channel']:
    shape = (n_geos, n_media_channels)
  elif dims == ['time', 'media_channel']:
    shape = (n_times, n_media_channels)
  else:
    raise ValueError(
        f'Shape {dims} not supported by the random_media_spend_nd_da function.'
    )

  media_spend = abs(np.random.normal(1, 1, size=shape))

  return xr.DataArray(
      media_spend,
      dims=dims,
      coords=coords,
      name='media_spend',
  )


def random_controls_da(
    media: xr.DataArray,
    n_geos: int,
    n_times: int,
    n_controls: int,
    seed: int = 0,
    date_format: str = c.DATE_FORMAT,
    explicit_time_index: Sequence[str] | None = None,
    integer_geos: bool = False,
) -> xr.DataArray:
  """Generates a sample `controls` DataArray.

  Args:
    media: The media data array
    n_geos: Number of geos
    n_times: Number of time periods
    n_controls: Number of controls
    seed: Random seed used by `np.random.seed()`
    date_format: The date format to use for time coordinate labels
    explicit_time_index: If given, ignore `date_format` and use this as is
    integer_geos: If True, the geos will be integers.

  Returns:
    A DataArray containing random controls variable.
  """

  np.random.seed(seed)

  controls_var = abs(np.random.normal(2, 1, size=(n_geos, n_times, 1)))
  media_common = media[:, (len(media.coords['media_time']) - n_times) :, :]
  controls = np.random.normal(
      0.8 * media_common.mean(axis=2, keepdims=True),
      controls_var,
      size=(n_geos, n_times, n_controls),
  )
  return xr.DataArray(
      controls,
      dims=['geo', 'time', 'control_variable'],
      coords={
          'geo': sample_geos(n_geos, integer_geos),
          'time': (
              _sample_times(n_times=n_times, date_format=date_format)
              if explicit_time_index is None
              else explicit_time_index
          ),
          'control_variable': _sample_names(
              prefix='control_', n_names=n_controls
          ),
      },
      name='controls',
  )


def random_kpi_da(
    media: xr.DataArray,
    controls: xr.DataArray,
    n_geos: int,
    n_times: int,
    n_media_channels: int,
    n_controls: int,
    seed: int = 0,
    integer_geos: bool = False,
) -> xr.DataArray:
  """Generates a sample `kpi` DataArray."""

  np.random.seed(seed)

  # Adds geo variance to media and geos to help model convergence.
  media_common = media[:, (len(media.coords['media_time']) - n_times) :, :]
  media_geo_sd = abs(np.random.normal(0, 5, size=n_geos))
  media_geo_sd = np.repeat(
      np.repeat(media_geo_sd[:, np.newaxis], n_times, axis=1)[..., np.newaxis],
      n_media_channels,
      axis=2,
  )
  control_geo_sd = abs(np.random.normal(0, 5, size=n_geos))
  control_geo_sd = np.repeat(
      np.repeat(control_geo_sd[:, np.newaxis], n_times, axis=1)[
          ..., np.newaxis
      ],
      n_controls,
      axis=2,
  )

  # Simulates outcome which is the dependent variable. Typically this is the
  # number of units sold, but it can be any metric (e.g. revenue).
  media_portion = np.random.normal(media_common, media_geo_sd).sum(axis=2)
  control_portion = np.random.normal(controls, control_geo_sd).sum(axis=2)
  error = np.random.normal(0, 2, size=(n_geos, n_times))
  kpi = abs(media_portion + control_portion + error)

  return xr.DataArray(
      kpi,
      dims=['geo', 'time'],
      coords={
          'geo': sample_geos(n_geos, integer_geos),
          'time': _sample_times(n_times=n_times),
      },
      name=c.KPI,
  )


def constant_revenue_per_kpi(
    n_geos: int,
    n_times: int,
    value: float,
    integer_geos: bool = False,
) -> xr.DataArray:
  """Generates a constant `revenue_per_kpi` DataArray."""

  revenue_per_kpi = value * np.ones((n_geos, n_times))

  return xr.DataArray(
      revenue_per_kpi,
      dims=['geo', 'time'],
      coords={
          'geo': sample_geos(n_geos, integer_geos),
          'time': _sample_times(n_times=n_times),
      },
      name='revenue_per_kpi',
  )


def random_population(
    n_geos: int, seed: int = 0, integer_geos: bool = False
) -> xr.DataArray:
  """Generates a sample `population` DataArray."""

  np.random.seed(seed)

  population = np.round(10 + abs(np.random.normal(3000, 100, size=n_geos)))

  return xr.DataArray(
      population,
      dims=['geo'],
      coords={'geo': sample_geos(n_geos, integer_geos)},
      name='population',
  )


def random_reach_da(
    n_geos: int,
    n_times: int,
    n_media_times: int,
    n_rf_channels: int,
    seed: int = 0,
    explicit_rf_channel_names: Sequence[str] | None = None,
    array_name: str = 'reach',
    channel_variable_name: str = 'rf_channel',
    channel_prefix: str = 'rf_ch_',
    integer_geos: bool = False,
) -> xr.DataArray:
  """Generates a sample `reach` DataArray."""

  np.random.seed(seed)

  start_date = _SAMPLE_START_DATE
  if n_times < n_media_times:
    start_date -= datetime.timedelta(weeks=(n_media_times - n_times))

  reach = np.round(
      abs(
          np.random.normal(
              3000, 100, size=(n_geos, n_media_times, n_rf_channels)
          )
      )
  )

  channels = (
      explicit_rf_channel_names
      if explicit_rf_channel_names is not None
      else _sample_names(prefix=channel_prefix, n_names=n_rf_channels)
  )
  return xr.DataArray(
      reach,
      dims=['geo', 'media_time', channel_variable_name],
      coords={
          'geo': sample_geos(n_geos, integer_geos),
          'media_time': _sample_times(
              n_times=n_media_times, start_date=start_date
          ),
          channel_variable_name: channels,
      },
      name=array_name,
  )


def random_organic_reach_da(
    n_geos: int,
    n_times: int,
    n_media_times: int,
    n_organic_rf_channels: int,
    seed: int = 0,
    explicit_organic_rf_channel_names: Sequence[str] | None = None,
    integer_geos: bool = False,
) -> xr.DataArray:
  """Generates a sample `organic_reach` DataArray."""
  return random_reach_da(
      n_geos=n_geos,
      n_times=n_times,
      n_media_times=n_media_times,
      n_rf_channels=n_organic_rf_channels,
      seed=seed,
      explicit_rf_channel_names=explicit_organic_rf_channel_names,
      array_name='organic_reach',
      channel_variable_name='organic_rf_channel',
      channel_prefix='organic_rf_ch_',
      integer_geos=integer_geos,
  )


def random_frequency_da(
    n_geos: int,
    n_times: int,
    n_media_times: int,
    n_rf_channels: int,
    seed: int = 0,
    explicit_rf_channel_names: Sequence[str] | None = None,
    array_name: str = 'frequency',
    channel_variable_name: str = 'rf_channel',
    channel_prefix: str = 'rf_ch_',
    integer_geos: bool = False,
) -> xr.DataArray:
  """Generates a sample `frequency` DataArray."""

  np.random.seed(seed)

  start_date = _SAMPLE_START_DATE
  if n_times < n_media_times:
    start_date -= datetime.timedelta(weeks=(n_media_times - n_times))

  frequency = abs(
      np.random.normal(3, 5, size=(n_geos, n_media_times, n_rf_channels))
  )

  channels = (
      explicit_rf_channel_names
      if explicit_rf_channel_names is not None
      else _sample_names(prefix=channel_prefix, n_names=n_rf_channels)
  )

  return xr.DataArray(
      frequency,
      dims=['geo', 'media_time', channel_variable_name],
      coords={
          'geo': sample_geos(n_geos, integer_geos),
          'media_time': _sample_times(
              n_times=n_media_times, start_date=start_date
          ),
          channel_variable_name: channels,
      },
      name=array_name,
  )


def random_organic_frequency_da(
    n_geos: int,
    n_times: int,
    n_media_times: int,
    n_organic_rf_channels: int,
    seed: int = 0,
    explicit_organic_rf_channel_names: Sequence[str] | None = None,
    integer_geos: bool = False,
) -> xr.DataArray:
  """Generates a sample `organic_frequency` DataArray."""
  return random_frequency_da(
      n_geos=n_geos,
      n_times=n_times,
      n_media_times=n_media_times,
      n_rf_channels=n_organic_rf_channels,
      seed=seed,
      explicit_rf_channel_names=explicit_organic_rf_channel_names,
      array_name='organic_frequency',
      channel_variable_name='organic_rf_channel',
      channel_prefix='organic_rf_ch_',
      integer_geos=integer_geos,
  )


def random_rf_spend_nd_da(
    n_geos: int | None = None,
    n_times: int | None = None,
    n_rf_channels: int | None = None,
    seed=0,
    integer_geos: bool = False,
) -> xr.DataArray:
  """Generates a sample N-dimensional `rf_spend` DataArray.

  This function generates a 1-D, 2-D or 3-D version of the `rf_spend` DataArray
  depending on the `n_geos`, `n_times` and `n_rf_channels` arguments.
  There are 3 accepted shapes accepted by the `InputData` class:
  `(rf_channel)`, `(geo, time, rf_channel)` and `(geo, rf_channel)`.

  Args:
    n_geos: Number of geos in the created `rf_spend` array or `None` if the
      created `rf_spend` should not contain the `geo` dimension.
    n_times: Number of time periods in the created `rf_spend` array or `None` if
      the created array should not contain the `time` dimension.
    n_rf_channels: Number of channels in the created `rf_spend` array.
    seed: Random seed used by `np.random.seed()`.
    integer_geos: If True, the geos will be integers.

  Returns:
    A DataArray containing the generated `rf_spend` data with the given
    dimensions.
  """
  np.random.seed(seed)

  dims = []
  coords = {}
  if n_geos is not None:
    dims.append('geo')
    coords['geo'] = sample_geos(n_geos, integer_geos)
  if n_times is not None:
    dims.append('time')
    coords['time'] = _sample_times(n_times=n_times)
  if n_rf_channels is not None:
    dims.append('rf_channel')
    coords['rf_channel'] = _sample_names(prefix='rf_ch_', n_names=n_rf_channels)

  if dims == ['geo', 'time', 'rf_channel']:
    shape = (n_geos, n_times, n_rf_channels)
  elif dims == ['rf_channel']:
    shape = (n_rf_channels,)
  elif dims == ['geo', 'rf_channel']:
    shape = (n_geos, n_rf_channels)
  else:
    raise ValueError(
        f'Shape {dims} not supported by the random_rf_spend_nd_da function.'
    )

  rf_spend = abs(np.random.normal(1, 1, size=shape))

  return xr.DataArray(
      rf_spend,
      dims=dims,
      coords=coords,
      name='rf_spend',
  )


def random_non_media_treatments_da(
    media: xr.DataArray,
    n_geos: int,
    n_times: int,
    n_non_media_channels: int,
    seed: int = 0,
    date_format: str = c.DATE_FORMAT,
    explicit_time_index: Sequence[str] | None = None,
    integer_geos: bool = False,
) -> xr.DataArray:
  """Generates a sample `non_media_treatments` DataArray.

  Args:
    media: The media data array
    n_geos: Number of geos
    n_times: Number of time periods
    n_non_media_channels: Number of non media channels
    seed: Random seed used by `np.random.seed()`
    date_format: The date format to use for time coordinate labels
    explicit_time_index: If given, ignore `date_format` and use this as is
    integer_geos: If True, the geos will be integers.

  Returns:
    A DataArray containing random non-media variable.
  """

  np.random.seed(seed)

  non_media_channel = abs(np.random.normal(2, 1, size=(n_geos, n_times, 1)))
  media_common = media[:, (len(media.coords['media_time']) - n_times) :, :]
  non_media_treatments = np.random.normal(
      0.8 * media_common.mean(axis=2, keepdims=True),
      non_media_channel,
      size=(n_geos, n_times, n_non_media_channels),
  )
  return xr.DataArray(
      non_media_treatments,
      dims=['geo', 'time', 'non_media_channel'],
      coords={
          'geo': sample_geos(n_geos, integer_geos),
          'time': (
              _sample_times(n_times=n_times, date_format=date_format)
              if explicit_time_index is None
              else explicit_time_index
          ),
          'non_media_channel': _sample_names(
              prefix='non_media_', n_names=n_non_media_channels
          ),
      },
      name='non_media_treatments',
  )


def random_dataset(
    n_geos: int,
    n_times: int,
    n_media_times: int,
    n_controls: int,
    n_non_media_channels: int | None = None,
    n_organic_media_channels: int | None = None,
    n_organic_rf_channels: int | None = None,
    n_media_channels: int | None = None,
    n_rf_channels: int | None = None,
    revenue_per_kpi_value: float | None = 3.14,
    seed: int = 0,
    remove_media_time: bool = False,
    integer_geos: bool = False,
):
  """Generates a random dataset."""
  if n_media_channels:
    media = random_media_da(
        n_geos=n_geos,
        n_times=n_times,
        n_media_times=n_media_times,
        n_media_channels=n_media_channels,
        seed=seed,
        integer_geos=integer_geos,
    )
    media_spend = random_media_spend_nd_da(
        n_geos=n_geos,
        n_times=n_times,
        n_media_channels=n_media_channels,
        seed=seed,
        integer_geos=integer_geos,
    )
  else:
    media = None
    media_spend = None

  if n_rf_channels:
    reach = random_reach_da(
        n_geos=n_geos,
        n_times=n_times,
        n_media_times=n_media_times,
        n_rf_channels=n_rf_channels,
        seed=seed,
        integer_geos=integer_geos,
    )
    frequency = random_frequency_da(
        n_geos=n_geos,
        n_times=n_times,
        n_media_times=n_media_times,
        n_rf_channels=n_rf_channels,
        seed=seed,
        integer_geos=integer_geos,
    )
    rf_spend = random_rf_spend_nd_da(
        n_geos=n_geos,
        n_times=n_times,
        n_rf_channels=n_rf_channels,
        seed=seed,
        integer_geos=integer_geos,
    )
  else:
    reach = None
    frequency = None
    rf_spend = None

  if revenue_per_kpi_value is not None:
    revenue_per_kpi = constant_revenue_per_kpi(
        n_geos=n_geos,
        n_times=n_times,
        value=revenue_per_kpi_value,
        integer_geos=integer_geos,
    )
  else:
    revenue_per_kpi = None

  controls = random_controls_da(
      media=media if n_media_channels else reach,
      n_geos=n_geos,
      n_times=n_times,
      n_controls=n_controls,
      seed=seed,
      integer_geos=integer_geos,
  )
  if n_non_media_channels:
    non_media_treatments = random_non_media_treatments_da(
        media=media if n_media_channels else reach,
        n_geos=n_geos,
        n_times=n_times,
        n_non_media_channels=n_non_media_channels,
        seed=seed,
        integer_geos=integer_geos,
    )
  else:
    non_media_treatments = None

  if n_organic_media_channels:
    organic_media = random_organic_media_da(
        n_geos=n_geos,
        n_times=n_times,
        n_media_times=n_media_times,
        n_organic_media_channels=n_organic_media_channels,
        seed=seed,
        integer_geos=integer_geos,
    )
  else:
    organic_media = None

  if n_organic_rf_channels:
    organic_reach = random_organic_reach_da(
        n_geos=n_geos,
        n_times=n_times,
        n_media_times=n_media_times,
        n_organic_rf_channels=n_organic_rf_channels,
        seed=seed,
        integer_geos=integer_geos,
    )
    organic_frequency = random_organic_frequency_da(
        n_geos=n_geos,
        n_times=n_times,
        n_media_times=n_media_times,
        n_organic_rf_channels=n_organic_rf_channels,
        seed=seed,
        integer_geos=integer_geos,
    )
  else:
    organic_reach = None
    organic_frequency = None

  kpi = random_kpi_da(
      media=media if n_media_channels else reach,
      controls=controls,
      n_geos=n_geos,
      n_times=n_times,
      n_media_channels=n_media_channels or n_rf_channels or 0,
      n_controls=n_controls,
      integer_geos=integer_geos,
  )
  population = random_population(
      n_geos=n_geos, seed=seed, integer_geos=integer_geos
  )

  dataset = xr.combine_by_coords([kpi, population, controls])
  if revenue_per_kpi is not None:
    dataset = xr.combine_by_coords([dataset, revenue_per_kpi])
  if media is not None:
    media_renamed = (
        media.rename({'media_time': 'time'}) if remove_media_time else media
    )

    dataset = xr.combine_by_coords([dataset, media_renamed, media_spend])
  if reach is not None:
    reach_renamed = (
        reach.rename({'media_time': 'time'}) if remove_media_time else reach
    )
    frequency_renamed = (
        frequency.rename({'media_time': 'time'})
        if remove_media_time
        else frequency
    )
    dataset = xr.combine_by_coords(
        [dataset, reach_renamed, frequency_renamed, rf_spend]
    )
  if non_media_treatments is not None:
    dataset = xr.combine_by_coords([dataset, non_media_treatments])
  if organic_media is not None:
    organic_media_renamed = (
        organic_media.rename({'media_time': 'time'})
        if remove_media_time
        else organic_media
    )
    dataset = xr.combine_by_coords([dataset, organic_media_renamed])
  if organic_reach is not None:
    organic_reach_renamed = (
        organic_reach.rename({'media_time': 'time'})
        if remove_media_time
        else organic_reach
    )
    organic_frequency_renamed = (
        organic_frequency.rename({'media_time': 'time'})
        if remove_media_time
        else organic_frequency
    )
    dataset = xr.combine_by_coords(
        [dataset, organic_reach_renamed, organic_frequency_renamed]
    )
  return dataset


def dataset_to_dataframe(
    dataset: xr.Dataset,
    controls_column_names: list[str],
    media_column_names: list[str] | None = None,
    media_spend_column_names: list[str] | None = None,
    reach_column_names: list[str] | None = None,
    frequency_column_names: list[str] | None = None,
    rf_spend_column_names: list[str] | None = None,
    non_media_column_names: list[str] | None = None,
    organic_media_column_names: list[str] | None = None,
    organic_reach_column_names: list[str] | None = None,
    organic_frequency_column_names: list[str] | None = None,
) -> pd.DataFrame:
  """Converts a dataset into the `pd.DataFrame` format.

  Args:
    dataset: An `xarray.Dataset` consisting of the following DataArrays: `kpi`,
      `revenue_per_kpi`, `population`, `media`, `media_spend`, and `controls`.
    controls_column_names: A list of desired column names for controls data in
      the output DataFrame.
    media_column_names: A list of desired column names for media data in the
      output DataFrame.
    media_spend_column_names: A list of desired column names for media_spend
      data in the output DataFrame.
    reach_column_names: A list of desired column names for reach data in the
      output DataFrame.
    frequency_column_names: A list of desired column names for frequency data in
      the output DataFrame.
    rf_spend_column_names: A list of desired column names for `rf_spend` data in
      the output DataFrame.
    non_media_column_names: A list of desired column names for
      non_media_treatments data in the output DataFrame.
    organic_media_column_names: A list of desired column names for organic_media
      data in the output DataFrame.
    organic_reach_column_names: A list of desired column names for organic_reach
      data in the output DataFrame.
    organic_frequency_column_names: A list of desired column names for
      organic_frequency data in the output DataFrame.

  Returns:
    A Pandas DataFrame with columns `geo` and `time` followed by `kpi`,
    `revenue_per_kpi`, `population`, `controls`, `media`, `media_spend`,
    `reach`, `frequency`, `rf_spend`, `non_media_treatments`, `organic media`,
    `organic_reach`, and `organic_frequency` data.
  """
  kpi = dataset[c.KPI].to_dataframe(name=c.KPI)
  revenue_per_kpi = dataset[c.REVENUE_PER_KPI].to_dataframe(
      name=c.REVENUE_PER_KPI
  )
  population = dataset[c.POPULATION].to_dataframe(name=c.POPULATION)

  controls = dataset[c.CONTROLS].to_dataframe(name=c.CONTROLS).unstack()
  controls.columns = controls_column_names

  result = kpi.join(revenue_per_kpi).join(population).join(controls)

  if non_media_column_names is not None:
    non_media_treatments = (
        dataset[c.NON_MEDIA_TREATMENTS]
        .to_dataframe(name=c.NON_MEDIA_TREATMENTS)
        .unstack()
    )
    non_media_treatments.columns = non_media_column_names
    result = result.join(non_media_treatments, how='right')

  if media_column_names is not None:
    media = dataset[c.MEDIA].to_dataframe(name=c.MEDIA).unstack()
    media.columns = media_column_names
    media.index.names = [c.GEO, c.TIME]

    media_spend = (
        dataset[c.MEDIA_SPEND].to_dataframe(name=c.MEDIA_SPEND).unstack()
    )
    media_spend.columns = media_spend_column_names

    result = result.join(media, how='right').join(media_spend)

  if reach_column_names is not None:
    reach = dataset[c.REACH].to_dataframe(name=c.REACH).unstack()
    reach.columns = reach_column_names
    reach.index.names = [c.GEO, c.TIME]

    frequency = dataset[c.FREQUENCY].to_dataframe(name=c.FREQUENCY).unstack()
    frequency.columns = frequency_column_names
    frequency.index.names = [c.GEO, c.TIME]

    rf_spend = dataset[c.RF_SPEND].to_dataframe(name=c.RF_SPEND).unstack()
    rf_spend.columns = rf_spend_column_names

    result = (
        result.join(reach, how='right')
        .join(frequency, how='right')
        .join(rf_spend)
    )

  if organic_media_column_names is not None:
    organic_media = (
        dataset[c.ORGANIC_MEDIA].to_dataframe(name=c.ORGANIC_MEDIA).unstack()
    )
    organic_media.columns = organic_media_column_names
    organic_media.index.names = [c.GEO, c.TIME]
    result = result.join(organic_media, how='right')

  if organic_reach_column_names is not None:
    organic_reach = (
        dataset[c.ORGANIC_REACH].to_dataframe(name=c.ORGANIC_REACH).unstack()
    )
    organic_reach.columns = organic_reach_column_names
    organic_reach.index.names = [c.GEO, c.TIME]

    organic_frequency = (
        dataset[c.ORGANIC_FREQUENCY]
        .to_dataframe(name=c.ORGANIC_FREQUENCY)
        .unstack()
    )
    organic_frequency.columns = organic_frequency_column_names
    organic_frequency.index.names = [c.GEO, c.TIME]
    result = result.join(organic_reach, how='right').join(organic_frequency)

  return result.reset_index()


def random_dataframe(
    n_geos,
    n_times,
    n_media_times,
    n_controls,
    n_media_channels=None,
    n_rf_channels=None,
    seed=0,
):
  """Generates a DataFrame for a random dataset."""
  dataset = random_dataset(
      n_geos=n_geos,
      n_times=n_times,
      n_media_times=n_media_times,
      n_controls=n_controls,
      n_media_channels=n_media_channels,
      n_rf_channels=n_rf_channels,
      seed=seed,
  )

  return dataset_to_dataframe(
      dataset,
      controls_column_names=_sample_names('control_', n_controls),
      media_column_names=_sample_names('media_', n_media_channels),
      media_spend_column_names=_sample_names('media_spend_', n_media_channels),
      reach_column_names=_sample_names('reach_', n_rf_channels),
      frequency_column_names=_sample_names('frequency_', n_rf_channels),
      rf_spend_column_names=_sample_names('rf_spend_', n_rf_channels),
  )


def sample_coord_to_columns(
    n_controls: int,
    n_media_channels: int | None = None,
    n_rf_channels: int | None = None,
    n_non_media_channels: int | None = None,
    n_organic_media_channels: int | None = None,
    n_organic_rf_channels: int | None = None,
    include_revenue_per_kpi: bool = True,
) -> load.CoordToColumns:
  """Returns a sample `coord_to_columns` mapping for testing."""

  if n_media_channels is not None:
    media = _sample_names('media_', n_media_channels)
    media_spend = _sample_names('media_spend_', n_media_channels)
  else:
    media = None
    media_spend = None

  if n_rf_channels is not None:
    reach = _sample_names('reach_', n_rf_channels)
    frequency = _sample_names('frequency_', n_rf_channels)
    rf_spend = _sample_names('rf_spend_', n_rf_channels)
  else:
    reach = None
    frequency = None
    rf_spend = None

  if n_organic_media_channels is not None:
    organic_media = _sample_names('organic_media_', n_organic_media_channels)
  else:
    organic_media = None

  if n_organic_rf_channels is not None:
    organic_reach = _sample_names('organic_reach_', n_organic_rf_channels)
    organic_frequency = _sample_names(
        'organic_frequency_', n_organic_rf_channels
    )
  else:
    organic_reach = None
    organic_frequency = None

  if n_non_media_channels is not None:
    non_media_treatments = _sample_names('non_media_', n_non_media_channels)
  else:
    non_media_treatments = None

  return load.CoordToColumns(
      geo=c.GEO,
      time=c.TIME,
      kpi=c.KPI,
      revenue_per_kpi=c.REVENUE_PER_KPI if include_revenue_per_kpi else None,
      population=c.POPULATION,
      controls=_sample_names('control_', n_controls),
      media=media,
      media_spend=media_spend,
      reach=reach,
      frequency=frequency,
      rf_spend=rf_spend,
      non_media_treatments=non_media_treatments,
      organic_media=organic_media,
      organic_reach=organic_reach,
      organic_frequency=organic_frequency,
  )


def sample_input_data_from_dataset(
    dataset: xr.Dataset, kpi_type: str
) -> input_data.InputData:
  """Generates a sample `InputData` from a full xarray Dataset."""
  return input_data.InputData(
      kpi=dataset.kpi,
      kpi_type=kpi_type,
      revenue_per_kpi=dataset.revenue_per_kpi,
      population=dataset.population,
      controls=dataset.controls,
      media=dataset.media,
      media_spend=dataset.media_spend,
      reach=dataset.reach,
      frequency=dataset.frequency,
      rf_spend=dataset.rf_spend,
  )


def sample_input_data_revenue(
    n_geos: int = 10,
    n_times: int = 50,
    n_media_times: int = 53,
    n_controls: int = 2,
    n_non_media_channels: int | None = None,
    n_media_channels: int | None = None,
    n_rf_channels: int | None = None,
    n_organic_media_channels: int | None = None,
    n_organic_rf_channels: int | None = None,
    seed: int = 0,
) -> input_data.InputData:
  """Generates sample InputData for `kpi_type='revenue'`."""
  dataset = random_dataset(
      n_geos=n_geos,
      n_times=n_times,
      n_media_times=n_media_times,
      n_controls=n_controls,
      n_non_media_channels=n_non_media_channels,
      n_media_channels=n_media_channels,
      n_rf_channels=n_rf_channels,
      n_organic_media_channels=n_organic_media_channels,
      n_organic_rf_channels=n_organic_rf_channels,
      revenue_per_kpi_value=1.0,
      seed=seed,
  )
  return input_data.InputData(
      kpi=dataset.kpi,
      kpi_type=c.REVENUE,
      revenue_per_kpi=None,
      population=dataset.population,
      controls=dataset.controls,
      non_media_treatments=dataset.non_media_treatments
      if n_non_media_channels
      else None,
      media=dataset.media if n_media_channels else None,
      media_spend=dataset.media_spend if n_media_channels else None,
      reach=dataset.reach if n_rf_channels else None,
      frequency=dataset.frequency if n_rf_channels else None,
      rf_spend=dataset.rf_spend if n_rf_channels else None,
      organic_media=dataset.organic_media if n_organic_media_channels else None,
      organic_reach=dataset.organic_reach if n_organic_rf_channels else None,
      organic_frequency=dataset.organic_frequency
      if n_organic_rf_channels
      else None,
  )


def sample_input_data_non_revenue_revenue_per_kpi(
    n_geos: int = 10,
    n_times: int = 50,
    n_media_times: int = 53,
    n_controls: int = 2,
    n_non_media_channels: int | None = None,
    n_media_channels: int | None = None,
    n_rf_channels: int | None = None,
    n_organic_media_channels: int | None = None,
    n_organic_rf_channels: int | None = None,
    seed: int = 0,
) -> input_data.InputData:
  """Generates sample InputData for `non_revenue` KPI w/ revenue_per_kpi."""
  dataset = random_dataset(
      n_geos=n_geos,
      n_times=n_times,
      n_media_times=n_media_times,
      n_controls=n_controls,
      n_non_media_channels=n_non_media_channels,
      n_media_channels=n_media_channels,
      n_rf_channels=n_rf_channels,
      n_organic_media_channels=n_organic_media_channels,
      n_organic_rf_channels=n_organic_rf_channels,
      seed=seed,
  )
  return input_data.InputData(
      kpi=dataset.kpi,
      kpi_type=c.NON_REVENUE,
      revenue_per_kpi=dataset.revenue_per_kpi,
      population=dataset.population,
      controls=dataset.controls,
      non_media_treatments=dataset.non_media_treatments
      if n_non_media_channels
      else None,
      media=dataset.media if n_media_channels else None,
      media_spend=dataset.media_spend if n_media_channels else None,
      reach=dataset.reach if n_rf_channels else None,
      frequency=dataset.frequency if n_rf_channels else None,
      rf_spend=dataset.rf_spend if n_rf_channels else None,
      organic_media=dataset.organic_media if n_organic_media_channels else None,
      organic_reach=dataset.organic_reach if n_organic_rf_channels else None,
      organic_frequency=dataset.organic_frequency
      if n_organic_rf_channels
      else None,
  )


def sample_input_data_non_revenue_no_revenue_per_kpi(
    n_geos: int = 10,
    n_times: int = 50,
    n_media_times: int = 53,
    n_controls: int = 2,
    n_non_media_channels: int | None = None,
    n_media_channels: int | None = None,
    n_rf_channels: int | None = None,
    n_organic_media_channels: int | None = None,
    n_organic_rf_channels: int | None = None,
    seed: int = 0,
) -> input_data.InputData:
  """Generates sample InputData for `non_revenue` KPI w/o revenue_per_kpi."""
  dataset = random_dataset(
      n_geos=n_geos,
      n_times=n_times,
      n_media_times=n_media_times,
      n_controls=n_controls,
      n_non_media_channels=n_non_media_channels,
      n_media_channels=n_media_channels,
      n_rf_channels=n_rf_channels,
      n_organic_media_channels=n_organic_media_channels,
      n_organic_rf_channels=n_organic_rf_channels,
      seed=seed,
  )
  return input_data.InputData(
      kpi=dataset.kpi,
      kpi_type=c.NON_REVENUE,
      revenue_per_kpi=None,
      population=dataset.population,
      controls=dataset.controls,
      non_media_treatments=dataset.non_media_treatments
      if n_non_media_channels
      else None,
      media=dataset.media if n_media_channels else None,
      media_spend=dataset.media_spend if n_media_channels else None,
      reach=dataset.reach if n_rf_channels else None,
      frequency=dataset.frequency if n_rf_channels else None,
      rf_spend=dataset.rf_spend if n_rf_channels else None,
      organic_media=dataset.organic_media if n_organic_media_channels else None,
      organic_reach=dataset.organic_reach if n_organic_rf_channels else None,
      organic_frequency=dataset.organic_frequency
      if n_organic_rf_channels
      else None,
  )



# Copyright 2024 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Summarization module that creates a 2-page HTML report."""

from collections.abc import Sequence
import functools
import os

import jinja2
from meridian import constants as c
from meridian.analysis import formatter
from meridian.analysis import summary_text
from meridian.analysis import visualizer
from meridian.data import time_coordinates as tc
from meridian.model import model
import pandas as pd
import xarray as xr


__all__ = [
    'Summarizer',
    'MODEL_FIT_CARD_SPEC',
    'CHANNEL_CONTRIB_CARD_SPEC',
    'PERFORMANCE_BREAKDOWN_CARD_SPEC',
    'RESPONSE_CURVES_CARD_SPEC',
]


MODEL_FIT_CARD_SPEC = formatter.CardSpec(
    id=summary_text.MODEL_FIT_CARD_ID,
    title=summary_text.MODEL_FIT_CARD_TITLE,
)
CHANNEL_CONTRIB_CARD_SPEC = formatter.CardSpec(
    id=summary_text.CHANNEL_CONTRIB_CARD_ID,
    title=summary_text.CHANNEL_CONTRIB_CARD_TITLE,
)
PERFORMANCE_BREAKDOWN_CARD_SPEC = formatter.CardSpec(
    id=summary_text.PERFORMANCE_BREAKDOWN_CARD_ID,
    title=summary_text.PERFORMANCE_BREAKDOWN_CARD_TITLE,
)
RESPONSE_CURVES_CARD_SPEC = formatter.CardSpec(
    id=summary_text.RESPONSE_CURVES_CARD_ID,
    title=summary_text.RESPONSE_CURVES_CARD_TITLE,
)


class Summarizer:
  """Generates HTML summary visualizations from the model fitting."""

  def __init__(self, meridian: model.Meridian):
    """Initialize the visualizer classes that are not time-dependent."""
    self._meridian = meridian

  @functools.cached_property
  def _model_fit(self):
    return visualizer.ModelFit(self._meridian)

  @functools.cached_property
  def _model_diagnostics(self):
    return visualizer.ModelDiagnostics(self._meridian)

  def output_model_results_summary(
      self,
      filename: str,
      filepath: str,
      start_date: tc.Date | None = None,
      end_date: tc.Date | None = None,
  ):
    """Generates and saves the HTML results summary output.

    Args:
      filename: The filename for the generated HTML output.
      filepath: The path to the directory where the file will be saved.
      start_date: Optional start date selector, *inclusive*, in _yyyy-mm-dd_
        format.
      end_date: Optional end date selector, *inclusive* in _yyyy-mm-dd_ format.
    """
    os.makedirs(filepath, exist_ok=True)
    with open(os.path.join(filepath, filename), 'w') as f:
      f.write(self._gen_model_results_summary(start_date, end_date))

  def _gen_model_results_summary(
      self,
      start_date: tc.Date | None = None,
      end_date: tc.Date | None = None,
  ) -> str:
    """Generate HTML results summary output (as sanitized content str)."""
    all_dates = self._meridian.input_data.time_coordinates.all_dates
    start_date = (
        tc.normalize_date(start_date)
        if start_date is not None
        else min(all_dates)
    )
    end_date = (
        tc.normalize_date(end_date) if end_date is not None else max(all_dates)
    )

    if start_date not in all_dates:
      raise ValueError(
          f'start_date ({start_date}) must be in the time coordinates!'
      )
    if end_date not in all_dates:
      raise ValueError(
          f'end_date ({end_date}) must be in the time coordinates!'
      )
    if start_date > end_date:
      raise ValueError(
          f'start_date ({start_date}) must be before end_date ({end_date})!'
      )

    selected_times = self._meridian.expand_selected_time_dims(
        start_date, end_date
    )

    template_env = formatter.create_template_env()
    template_env.globals[c.START_DATE] = start_date.strftime(
        f'%b {start_date.day}, %Y'
    )

    interval_days = self._meridian.input_data.time_coordinates.interval_days
    end_date_adjusted = end_date + pd.Timedelta(days=interval_days)

    template_env.globals[c.END_DATE] = end_date_adjusted.strftime(
        f'%b {end_date_adjusted.day}, %Y'
    )

    html_template = template_env.get_template('summary.html.jinja')
    cards_htmls = self._create_cards_htmls(
        template_env,
        selected_times=selected_times,
    )

    return html_template.render(
        title=summary_text.MODEL_RESULTS_TITLE, cards=cards_htmls
    )

  def _create_cards_htmls(
      self,
      template_env: jinja2.Environment,
      selected_times: Sequence[str] | None,
  ) -> Sequence[str]:
    """Creates the HTML snippets for cards in the summary page."""
    media_summary = visualizer.MediaSummary(
        self._meridian, selected_times=selected_times
    )
    media_effects = visualizer.MediaEffects(self._meridian)
    reach_frequency = (
        visualizer.ReachAndFrequency(
            self._meridian, selected_times=selected_times
        )
        if self._meridian.n_rf_channels > 0
        else None
    )
    cards = [
        self._create_model_fit_card_html(
            template_env, selected_times=selected_times
        ),
        self._create_outcome_contrib_card_html(template_env, media_summary),
        self._create_performance_breakdown_card_html(
            template_env, media_summary
        ),
        self._create_response_curves_card_html(
            template_env=template_env,
            selected_times=selected_times,
            media_summary=media_summary,
            media_effects=media_effects,
            reach_frequency=reach_frequency,
        ),
    ]
    return cards

  def _create_model_fit_card_html(
      self, template_env: jinja2.Environment, **kwargs
  ) -> str:
    """Creates the HTML snippet for the Model Fit card."""
    model_fit = self._model_fit
    outcome = self._kpi_or_revenue()
    expected_actual_outcome_chart = formatter.ChartSpec(
        id=summary_text.EXPECTED_ACTUAL_OUTCOME_CHART_ID,
        description=summary_text.EXPECTED_ACTUAL_OUTCOME_CHART_DESCRIPTION_FORMAT.format(
            outcome=outcome
        ),
        chart_json=model_fit.plot_model_fit(**kwargs).to_json(),
    )

    predictive_accuracy_table = self._predictive_accuracy_table_spec(**kwargs)
    insights = summary_text.MODEL_FIT_INSIGHTS_FORMAT

    return formatter.create_card_html(
        template_env,
        MODEL_FIT_CARD_SPEC,
        insights,
        [expected_actual_outcome_chart, predictive_accuracy_table],
    )

  def _predictive_accuracy_table_spec(self, **kwargs) -> formatter.TableSpec:
    """Creates the HTML snippet for the predictive accuracy table."""
    outcome = self._kpi_or_revenue()
    model_diag = self._model_diagnostics
    table = model_diag.predictive_accuracy_table(column_var=c.METRIC, **kwargs)

    # Only take the national stats, even if geo ones exist.
    national_table = table[table[c.GEO_GRANULARITY] == c.NATIONAL]

    # Translate column names into human-presentable ones.
    column_names = [
        summary_text.DATASET_LABEL,
        summary_text.R_SQUARED_LABEL,
        summary_text.MAPE_LABEL,
        summary_text.WMAPE_LABEL,
    ]

    if c.EVALUATION_SET_VAR in list(national_table.columns):

      def _slice_table_by_evaluation_set(eval_set: str) -> Sequence[str]:
        """Slices table by the given evaluation set."""
        sliced_table_by_eval_set = national_table[
            national_table[c.EVALUATION_SET_VAR] == eval_set
        ]
        row_values = [
            '{:.2f}'.format(sliced_table_by_eval_set[c.R_SQUARED].item()),
            formatter.format_percent(sliced_table_by_eval_set[c.MAPE].item()),
            formatter.format_percent(sliced_table_by_eval_set[c.WMAPE].item()),
        ]
        return row_values

      # The dataset has holdout_id that distinguish training and test data sets.
      training_row = [summary_text.TRAINING_DATA_LABEL]
      training_row.extend(_slice_table_by_evaluation_set(c.TRAIN))
      testing_row = [summary_text.TESTING_DATA_LABEL]
      testing_row.extend(_slice_table_by_evaluation_set(c.TEST))
      all_data_row = [summary_text.ALL_DATA_LABEL]
      all_data_row.extend(_slice_table_by_evaluation_set(c.ALL_DATA))

      row_values = [training_row, testing_row, all_data_row]
    else:  # No holdout_id present, so metrics are taken from 'All Data'.
      row_values = [[
          summary_text.ALL_DATA_LABEL,
          '{:.2f}'.format(national_table[c.R_SQUARED].item()),
          '{:.0%}'.format(national_table[c.MAPE].item()),
          '{:.0%}'.format(national_table[c.WMAPE].item()),
      ]]

    return formatter.TableSpec(
        id=summary_text.PREDICTIVE_ACCURACY_TABLE_ID,
        title=summary_text.PREDICTIVE_ACCURACY_TABLE_TITLE,
        description=summary_text.PREDICTIVE_ACCURACY_TABLE_DESCRIPTION.format(
            outcome=outcome
        ),
        column_headers=column_names,
        row_values=row_values,
    )

  def _create_outcome_contrib_card_html(
      self,
      template_env: jinja2.Environment,
      media_summary: visualizer.MediaSummary,
  ) -> str:
    """Creates the HTML snippet for the Outcome Contrib card."""
    outcome = self._kpi_or_revenue()
    channel_drivers_chart = formatter.ChartSpec(
        id=summary_text.CHANNEL_DRIVERS_CHART_ID,
        description=summary_text.CHANNEL_DRIVERS_CHART_DESCRIPTION.format(
            outcome=outcome
        ),
        chart_json=media_summary.plot_contribution_waterfall_chart().to_json(),
    )
    lead_channels = self._get_sorted_posterior_mean_metrics_df(
        media_summary, [c.INCREMENTAL_OUTCOME]
    )[c.CHANNEL][:2]
    formatted_channels = [channel.title() for channel in lead_channels]

    spend_outcome_chart = formatter.ChartSpec(
        id=summary_text.SPEND_OUTCOME_CHART_ID,
        description=summary_text.SPEND_OUTCOME_CHART_DESCRIPTION.format(
            outcome=outcome
        ),
        chart_json=media_summary.plot_spend_vs_contribution().to_json(),
    )
    outcome_contribution_chart = formatter.ChartSpec(
        id=summary_text.OUTCOME_CONTRIBUTION_CHART_ID,
        description=summary_text.OUTCOME_CONTRIBUTION_CHART_DESCRIPTION.format(
            outcome=outcome
        ),
        chart_json=media_summary.plot_contribution_pie_chart().to_json(),
    )
    insights = summary_text.CHANNEL_CONTRIB_INSIGHTS_FORMAT.format(
        outcome=outcome,
        lead_channels=' and '.join(formatted_channels),
    )
    return formatter.create_card_html(
        template_env,
        CHANNEL_CONTRIB_CARD_SPEC,
        insights,
        [
            channel_drivers_chart,
            spend_outcome_chart,
            outcome_contribution_chart,
        ],
    )

  def _get_sorted_posterior_mean_metrics_df(
      self,
      media_summary: visualizer.MediaSummary,
      metrics: Sequence[str],
      ascending: bool = False,
  ) -> pd.DataFrame:
    return (
        media_summary.paid_summary_metrics[metrics]
        .sel(distribution=c.POSTERIOR, metric=c.MEAN)
        .drop_sel(channel=c.ALL_CHANNELS)
        .to_dataframe()
        .drop(columns=[c.METRIC, c.DISTRIBUTION])
        .sort_values(by=metrics, ascending=ascending)
        .reset_index()
    )

  def _get_sorted_posterior_median_metrics_df(
      self,
      media_summary: visualizer.MediaSummary,
      metrics: Sequence[str],
      ascending: bool = False,
  ) -> pd.DataFrame:
    return (
        media_summary.paid_summary_metrics[metrics]
        .sel(distribution=c.POSTERIOR, metric=c.MEDIAN)
        .drop_sel(channel=c.ALL_CHANNELS)
        .to_dataframe()
        .drop(columns=[c.METRIC, c.DISTRIBUTION])
        .sort_values(by=metrics, ascending=ascending)
        .reset_index()
    )

  def _create_performance_breakdown_card_html(
      self,
      template_env: jinja2.Environment,
      media_summary: visualizer.MediaSummary,
  ) -> str:
    """Creates the HTML snippet for the ROI and CPIK Breakdown card."""
    roi_effectiveness_chart = formatter.ChartSpec(
        id=summary_text.ROI_EFFECTIVENESS_CHART_ID,
        description=summary_text.ROI_EFFECTIVENESS_CHART_DESCRIPTION,
        chart_json=media_summary.plot_roi_vs_effectiveness().to_json(),
    )
    roi_marginal_chart = formatter.ChartSpec(
        id=summary_text.ROI_MARGINAL_CHART_ID,
        description=summary_text.ROI_MARGINAL_CHART_DESCRIPTION,
        chart_json=media_summary.plot_roi_vs_mroi().to_json(),
    )
    roi_channel_chart = formatter.ChartSpec(
        id=summary_text.ROI_CHANNEL_CHART_ID,
        chart_json=media_summary.plot_roi_bar_chart().to_json(),
    )
    cpik_channel_chart = formatter.ChartSpec(
        id=summary_text.CPIK_CHANNEL_CHART_ID,
        chart_json=media_summary.plot_cpik().to_json(),
        description=summary_text.CPIK_CHANNEL_CHART_DESCRIPTION,
    )
    roi_df = self._get_sorted_posterior_mean_metrics_df(media_summary, [c.ROI])
    effectiveness_df = self._get_sorted_posterior_mean_metrics_df(
        media_summary, [c.EFFECTIVENESS]
    )
    mroi_df = self._get_sorted_posterior_mean_metrics_df(
        media_summary, [c.MROI]
    )
    cpik_df = self._get_sorted_posterior_median_metrics_df(
        media_summary, [c.CPIK], ascending=True
    )
    insights = summary_text.PERFORMANCE_BREAKDOWN_INSIGHTS_FORMAT.format(
        lead_roi_channel=roi_df[c.CHANNEL][0].title(),
        lead_roi_ratio=roi_df[c.ROI][0],
        lead_effectiveness_channel=effectiveness_df[c.CHANNEL][0].title(),
        lead_mroi_channel=mroi_df[c.CHANNEL][0].title(),
        lead_mroi_channel_value=mroi_df[c.MROI][0],
        lead_cpik_channel=cpik_df[c.CHANNEL][0].title(),
        lead_cpik_ratio=cpik_df[c.CPIK][0],
    )
    return formatter.create_card_html(
        template_env,
        PERFORMANCE_BREAKDOWN_CARD_SPEC,
        insights,
        [
            roi_effectiveness_chart,
            roi_marginal_chart,
            roi_channel_chart,
            cpik_channel_chart,
        ],
    )

  def _create_response_curves_card_html(
      self,
      template_env: jinja2.Environment,
      selected_times: Sequence[str] | None,
      media_summary: visualizer.MediaSummary,
      media_effects: visualizer.MediaEffects,
      reach_frequency: visualizer.ReachAndFrequency | None,
  ) -> str:
    """Creates the HTML snippet for the Optimal Analyst card."""
    outcome = self._kpi_or_revenue()
    charts = []
    charts.append(
        formatter.ChartSpec(
            id=summary_text.RESPONSE_CURVES_CHART_ID,
            description=summary_text.RESPONSE_CURVES_CHART_DESCRIPTION_FORMAT.format(
                outcome=outcome
            ),
            chart_json=media_effects.plot_response_curves(
                confidence_level=c.DEFAULT_CONFIDENCE_LEVEL,
                selected_times=(
                    frozenset(selected_times) if selected_times else None
                ),
                plot_separately=False,
                include_ci=False,
                num_channels_displayed=7,
            ).to_json(),
        )
    )

    insights = summary_text.RESPONSE_CURVES_INSIGHTS_FORMAT.format(
        outcome=outcome
    )
    if reach_frequency is not None:
      assert self._meridian.n_rf_channels > 0
      optimal_rf = self._select_optimal_rf_data(media_summary, reach_frequency)
      channel_name = optimal_rf[c.RF_CHANNEL].values.item()
      opt_freq = '{:.1f}'.format(optimal_rf.values.item())
      description = summary_text.OPTIMAL_FREQ_CHART_DESCRIPTION
      insights = ' '.join([
          insights,
          summary_text.OPTIMAL_FREQUENCY_INSIGHTS_FORMAT.format(
              rf_channel=channel_name,
              opt_freq=opt_freq,
          ),
      ])

      charts.append(
          formatter.ChartSpec(
              id=summary_text.OPTIMAL_FREQUENCY_CHART_ID,
              description=description,
              chart_json=reach_frequency.plot_optimal_frequency(
                  selected_channels=[channel_name],
              ).to_json(),
          )
      )

    return formatter.create_card_html(
        template_env, RESPONSE_CURVES_CARD_SPEC, insights, charts
    )

  def _select_optimal_rf_data(
      self,
      media_summary: visualizer.MediaSummary,
      reach_frequency: visualizer.ReachAndFrequency,
  ) -> xr.DataArray:
    """Selects and returns the `optimal_frequency` DataArray--if any.

    The `optimal_frequency` data is a subset of the Dataset property
    `optimal_frequency_data` of visualizer.ReachAndFrequency.

    Assumes that there is at least 1 RF channel in the model, else ValueError.
    Returns:
      DataArray of the optimal_frequency data for the channel with the highest
      spend value (per MediaSummary).
    """
    # Select the optimal frequency channel with the most spend.
    # This raises ValueError if there is no RF channel in the model.
    rf_channels = reach_frequency.optimal_frequency_data.rf_channel
    assert rf_channels.size > 0
    # This will raise KeyError if not all `rf_channels` can be found in here:
    rf_channel_spends = media_summary.paid_summary_metrics[c.SPEND].sel(
        channel=rf_channels
    )
    most_spend_rf_channel = rf_channel_spends.idxmax()

    return reach_frequency.optimal_frequency_data.sel(
        rf_channel=most_spend_rf_channel
    ).optimal_frequency

  def _kpi_or_revenue(self) -> str:
    if self._meridian.input_data.revenue_per_kpi is not None:
      outcome_str = c.REVENUE
    else:
      outcome_str = c.KPI.upper()
    return outcome_str
